# Generated by Django 3.2.15 on 2022-08-12 16:28

from django.db import migrations, models, transaction


def convert_artifact_to_snippets(apps, schema_editor):
    """Create snippet from artifact and remove artifact."""
    with transaction.atomic():
        Modulemd = apps.get_model('rpm', 'Modulemd')
        ModulemdDefaults = apps.get_model('rpm', 'ModulemdDefaults')
        ContentArtifact = apps.get_model('core', 'ContentArtifact')
        Artifact = apps.get_model('core', 'Artifact')
        modules_with_snippet = []
        defaults_with_snippet = []
        content_artifacts_to_delete = []
        artifacts_to_delete = []

        for module in Modulemd.objects.all():
            artifact = module._artifacts.get()
            module.snippet = artifact.file.read()
            content_artifact = ContentArtifact.objects.get(artifact__pk=artifact.pk)
            content_artifacts_to_delete.append(content_artifact.pk)
            artifacts_to_delete.append(artifact.pk)
            modules_with_snippet.append(module)

        for default in ModulemdDefaults.objects.all():
            artifact = default._artifacts.get()
            default.snippet = artifact.file.read()
            content_artifact = ContentArtifact.objects.get(artifact__pk=artifact.pk)
            content_artifacts_to_delete.append(content_artifact.pk)
            artifacts_to_delete.append(artifact.pk)
            defaults_with_snippet.append(default)

        Modulemd.objects.bulk_update(
            modules_with_snippet,
            ['snippet']
        )

        ModulemdDefaults.objects.bulk_update(
            defaults_with_snippet,
            ['snippet']
        )

        ContentArtifact.objects.filter(pk__in=content_artifacts_to_delete).delete()
        Artifact.objects.filter(pk__in=artifacts_to_delete).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0043_textfield_conversion'),
    ]

    operations = [
        migrations.AddField(
            model_name='modulemd',
            name='snippet',
            field=models.TextField(default='snippet'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='modulemddefaults',
            name='snippet',
            field=models.TextField(default='snippet'),
            preserve_default=False,
        ),
        migrations.RunPython(convert_artifact_to_snippets),
    ]



# Generated by Django 2.2.16 on 2020-10-09 15:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0025_remove_orphaned_subrepos'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmpublication',
            name='gpgcheck',
            field=models.IntegerField(choices=[(0, 0), (1, 1)], default=0),
        ),
        migrations.AddField(
            model_name='rpmpublication',
            name='repo_gpgcheck',
            field=models.IntegerField(choices=[(0, 0), (1, 1)], default=0),
        ),
    ]



"""
Check `Plugin Writer's Guide`_ for more details.

.. _Plugin Writer's Guide:
    http://docs.pulpproject.org/plugins/plugin-writer/index.html
"""

INSTALLED_APPS = ["django_readonly_field", "dynaconf_merge"]
ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION = False
DEFAULT_ULN_SERVER_BASE_URL = "https://linux-update.oracle.com/"
KEEP_CHANGELOG_LIMIT = 10
SOLVER_DEBUG_LOGS = True



#!/usr/bin/env python3

from setuptools import setup, find_packages

with open("requirements.txt") as requirements:
    requirements = requirements.readlines()

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="pulp-rpm",
    version="3.18.0",
    description="RPM plugin for the Pulp Project",
    long_description=long_description,
    license="GPLv2+",
    author="Pulp Project Developers",
    author_email="pulp-list@redhat.com",
    url="http://www.pulpproject.org",
    python_requires=">=3.8",
    install_requires=requirements,
    include_package_data=True,
    packages=find_packages(exclude=["test"]),
    classifiers=(
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ),
    entry_points={
        "pulpcore.plugin": [
            "pulp_rpm = pulp_rpm:default_app_config",
        ],
    },
)



"""
Tests comps.xml upload functionality.
"""
from tempfile import NamedTemporaryFile

from pulpcore.client.pulp_rpm import (
    RepositoriesRpmApi,
    RepositoriesRpmVersionsApi,
    ContentPackagegroupsApi,
    ContentPackageenvironmentsApi,
    ContentPackagelangpacksApi,
    ContentPackagecategoriesApi,
    RpmCompsApi,
)

from pulp_rpm.tests.functional.constants import (
    BIG_COMPS_XML,
    BIG_CATEGORY,
    BIG_GROUPS,
    BIG_ENVIRONMENTS,
    BIG_LANGPACK,
    RPM_PACKAGEENVIRONMENT_CONTENT_NAME,
    RPM_PACKAGECATEGORY_CONTENT_NAME,
    RPM_PACKAGEGROUP_CONTENT_NAME,
    RPM_PACKAGELANGPACKS_CONTENT_NAME,
    SMALL_COMPS_XML,
    SMALL_CATEGORY,
    SMALL_ENVIRONMENTS,
    SMALL_GROUPS,
    SMALL_LANGPACK,
)
from pulp_rpm.tests.functional.utils import (
    gen_rpm_client,
)

from pulp_smash import api, config
from pulp_smash.pulp3.bindings import (
    delete_orphans,
    monitor_task,
    PulpTestCase,
)
from pulp_smash.pulp3.utils import gen_repo


class CompsXmlTestCase(PulpTestCase):
    """
    Test comps.xml-upload functionality.
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.rpm_client = gen_rpm_client()
        cls.comps_api = RpmCompsApi(cls.rpm_client)
        cls.repo_api = RepositoriesRpmApi(cls.rpm_client)
        cls.repo_version_api = RepositoriesRpmVersionsApi(cls.rpm_client)
        cls.groups_api = ContentPackagegroupsApi(cls.rpm_client)
        cls.envs_api = ContentPackageenvironmentsApi(cls.rpm_client)
        cls.groupslangpacks_api = ContentPackagelangpacksApi(cls.rpm_client)
        cls.categories_api = ContentPackagecategoriesApi(cls.rpm_client)
        cls.small_content = SMALL_GROUPS + SMALL_CATEGORY + SMALL_LANGPACK + SMALL_ENVIRONMENTS
        cls.centos8_content = BIG_GROUPS + BIG_CATEGORY + BIG_LANGPACK + BIG_ENVIRONMENTS

    def _upload_comps_into(self, file_path, expected_totals, repo_href=None, replace=False):
        data = {}
        if repo_href:
            data["repository"] = repo_href
            data["replace"] = replace
            expected_totals += 1
        response = self.comps_api.rpm_comps_upload(file=file_path, **data)
        task = monitor_task(response.task)
        rsrcs = task.created_resources
        self.assertEqual(expected_totals, len(rsrcs))
        return rsrcs

    def _eval_resources(self, resources, is_small=True):
        """Eval created_resources counts."""
        groups = [g for g in resources if "packagegroups" in g]
        self.assertEqual(len(groups), SMALL_GROUPS if is_small else BIG_GROUPS)

        categories = [g for g in resources if "packagecategories" in g]
        self.assertEqual(len(categories), SMALL_CATEGORY if is_small else BIG_CATEGORY)

        langpacks = [g for g in resources if "packagelangpacks" in g]
        self.assertEqual(len(langpacks), SMALL_LANGPACK if is_small else BIG_LANGPACK)

        envs = [g for g in resources if "environment" in g]
        self.assertEqual(len(envs), SMALL_ENVIRONMENTS if is_small else BIG_ENVIRONMENTS)

    def _eval_counts(self, summary, is_small=True):
        """Eval counts in a given summary."""
        for c in summary:
            if "rpm.packagegroup" in c:
                self.assertEquals(
                    summary["rpm.packagegroup"]["count"], SMALL_GROUPS if is_small else BIG_GROUPS
                )
            elif "rpm.packageenvironment" in c:
                self.assertEquals(
                    summary["rpm.packageenvironment"]["count"],
                    SMALL_ENVIRONMENTS if is_small else BIG_ENVIRONMENTS,
                )
            elif RPM_PACKAGECATEGORY_CONTENT_NAME in c:
                self.assertEquals(
                    summary[RPM_PACKAGECATEGORY_CONTENT_NAME]["count"],
                    SMALL_CATEGORY if is_small else BIG_CATEGORY,
                )
            elif "rpm.packagelangpacks" in c:
                self.assertEquals(
                    summary["rpm.packagelangpacks"]["count"],
                    SMALL_LANGPACK if is_small else BIG_LANGPACK,
                )

    def _eval_sum_counts(self, summary):
        """In a given summary, counts should be BIG+SMALL."""
        for c in summary:
            if RPM_PACKAGEGROUP_CONTENT_NAME in c:
                self.assertEquals(
                    summary[RPM_PACKAGEGROUP_CONTENT_NAME]["count"], (BIG_GROUPS + SMALL_GROUPS)
                )
            elif RPM_PACKAGEENVIRONMENT_CONTENT_NAME in c:
                self.assertEquals(
                    summary[RPM_PACKAGEENVIRONMENT_CONTENT_NAME]["count"],
                    (BIG_ENVIRONMENTS + SMALL_ENVIRONMENTS),
                )
            elif RPM_PACKAGECATEGORY_CONTENT_NAME in c:
                self.assertEquals(
                    summary[RPM_PACKAGECATEGORY_CONTENT_NAME]["count"],
                    (BIG_CATEGORY + SMALL_CATEGORY),
                )
            elif RPM_PACKAGELANGPACKS_CONTENT_NAME in c:
                self.assertEquals(
                    summary[RPM_PACKAGELANGPACKS_CONTENT_NAME]["count"],
                    (BIG_LANGPACK + SMALL_LANGPACK),
                )

    def setUp(self) -> None:
        """Set up a repo for us to upload into."""
        self.repo = self.repo_api.create(gen_repo())

    def tearDown(self) -> None:
        """Throw away the test repo and content."""
        self.repo_api.delete(self.repo.pulp_href)
        delete_orphans()

    def test_upload(self):
        """Upload a comps.xml and make sure it created comps-Content."""
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(SMALL_COMPS_XML)
            comps_file.flush()
            resources = self._upload_comps_into(comps_file.name, self.small_content)
        self._eval_resources(resources)

    def test_upload_same(self):
        """Upload a comps.xml twice and make sure it doesn't create new objects the second time."""
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(SMALL_COMPS_XML)
            comps_file.flush()
            first = self._upload_comps_into(comps_file.name, self.small_content)
            second = self._upload_comps_into(comps_file.name, self.small_content)

        # we return all resources in the comps.xml, even if they already existed
        self.assertEqual(sorted(first), sorted(second))

    def test_upload_diff(self):
        """Upload two different comps-files and make sure the second shows the new comps-Content."""
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(SMALL_COMPS_XML)
            comps_file.flush()
            self._upload_comps_into(comps_file.name, self.small_content)
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(BIG_COMPS_XML)
            comps_file.flush()
            second = self._upload_comps_into(comps_file.name, self.centos8_content)
        self._eval_resources(second, is_small=False)

    def test_upload_into_repo(self):
        """Upload comps into a repo and see new version created containing the comps-Content."""
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(SMALL_COMPS_XML)
            comps_file.flush()
            resources = self._upload_comps_into(
                comps_file.name, self.small_content, self.repo.pulp_href
            )

        vers = [g for g in resources if "versions" in g]
        self.assertEqual(len(vers), 1)
        vers_resp = self.repo_version_api.read(vers[0])
        self.assertEqual(len(vers_resp.content_summary.added), 3)
        self.assertEqual(len(vers_resp.content_summary.present), 3)
        self._eval_counts(vers_resp.content_summary.added)

    def test_upload_into_repo_add(self):
        """Upload two comps-files into a repo and see the result being additive."""
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(SMALL_COMPS_XML)
            comps_file.flush()
            resources = self._upload_comps_into(
                comps_file.name, self.small_content, self.repo.pulp_href
            )

        vers = [g for g in resources if "versions" in g]
        self.assertEqual(len(vers), 1)
        vers_resp = self.repo_version_api.read(vers[0])
        self.assertEqual(len(vers_resp.content_summary.added), 3)
        self.assertEqual(len(vers_resp.content_summary.present), 3)

        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(BIG_COMPS_XML)
            comps_file.flush()
            resources = self._upload_comps_into(
                comps_file.name, self.centos8_content, self.repo.pulp_href
            )
        vers = [g for g in resources if "versions" in g]
        self.assertEqual(len(vers), 1)
        vers_resp = self.repo_version_api.read(vers[0])
        self.assertEqual(vers_resp.number, 2)

        self.assertEqual(len(vers_resp.content_summary.added), 3)
        self.assertEqual(len(vers_resp.content_summary.present), 4)
        self._eval_counts(vers_resp.content_summary.added, is_small=False)
        self._eval_sum_counts(vers_resp.content_summary.present)

    def test_upload_into_repo_replace(self):
        """Upload two comps, see the comps-content from the second replace the existing."""
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(SMALL_COMPS_XML)
            comps_file.flush()
            self._upload_comps_into(comps_file.name, self.small_content, self.repo.pulp_href)
        with NamedTemporaryFile("w+") as comps_file:
            comps_file.write(BIG_COMPS_XML)
            comps_file.flush()
            resources = self._upload_comps_into(
                comps_file.name, self.centos8_content, self.repo.pulp_href, True
            )
        vers = [g for g in resources if "versions" in g]
        self.assertEqual(len(vers), 1)
        vers_resp = self.repo_version_api.read(vers[0])
        self.assertEqual(vers_resp.number, 2)

        self.assertEqual(len(vers_resp.content_summary.added), 3)
        self.assertEqual(len(vers_resp.content_summary.present), 3)
        self._eval_counts(vers_resp.content_summary.added, is_small=False)



# Generated by Django 3.2.15 on 2022-08-29 16:33

from django.db import migrations, models, transaction

import gi
import logging

gi.require_version("Modulemd", "2.0")
from gi.repository import Modulemd as mmdlib  # noqa: E402


logger = logging.getLogger(__name__)


def populate_new_fields(apps, schema_editor):
    """Populate new fields of modulemd."""
    Modulemd = apps.get_model("rpm", "Modulemd")
    modules_to_update = []

    for modulemd in Modulemd.objects.filter(profiles={}):
        module_index = mmdlib.ModuleIndex.new()
        try:
            ret, err = module_index.update_from_string(modulemd.snippet, False)
            if err:
                raise ValueError(
                    "Repair failed on module "
                    f"'{modulemd.name}.{modulemd.stream}.{modulemd.version}'."
                )
            for module in module_index.get_module_names():
                for stream in module_index.get_module(module).get_all_streams():
                    modulemd.profiles = {}
                    modulemd.description = stream.get_description()
                    for profile in stream.get_profile_names():
                        stream_profile = stream.get_profile(profile)
                        modulemd.profiles[profile] = stream_profile.get_rpms()
                        modules_to_update.append(modulemd)
        except ValueError as err:
            # Due to issue #2735 it could happen that snippet will be empty
            # https://github.com/pulp/pulp_rpm/issues/2735
            logger.warning(
                "Modulemd {}-{}-{} cannot populate new fields."
                "So it will miss the info about profiles and its description.".format(
                    modulemd.name, modulemd.stream, modulemd.version
                )
            )

    Modulemd.objects.bulk_update(modules_to_update, ["profiles", "description"])


class Migration(migrations.Migration):

    dependencies = [
        ("rpm", "0044_noartifact_modules"),
    ]

    operations = [
        migrations.AddField(
            model_name="modulemd",
            name="description",
            field=models.TextField(default="Description"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="modulemd",
            name="profiles",
            field=models.JSONField(default=dict),
        ),
        migrations.RunPython(populate_new_fields),
    ]



"""Tests that publish rpm plugin repositories."""
import unittest
from datetime import datetime
from html.parser import HTMLParser
from tempfile import NamedTemporaryFile
from urllib.parse import urljoin

from productmd.treeinfo import TreeInfo

from pulp_smash import api, config, utils
from pulp_smash.pulp3.bindings import delete_orphans
from pulp_smash.pulp3.utils import (
    gen_repo,
    get_added_content_summary,
    get_content_summary,
)

from pulp_rpm.tests.functional.constants import (
    RPM_DISTRIBUTION_PATH,
    RPM_PACKAGE_CONTENT_NAME,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_PUBLICATION_PATH,
    RPM_REMOTE_PATH,
    RPM_REPO_PATH,
    CENTOS7_URL,
    CENTOS8_STREAM_APPSTREAM_URL,
    CENTOS8_STREAM_BASEOS_URL,
    EPEL7_URL,
)
from pulp_rpm.tests.functional.utils import gen_rpm_remote
from pulp_rpm.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class PackagesHtmlParser(HTMLParser):
    """HTML parser that looks for the first link to a file.

    This parser look for the first link that doesn't have a trailing slash. It then sets the
    package_href attribute to the value of the href.
    """

    package_href = None

    def handle_starttag(self, tag, attrs):
        """Looks for a tags that don't end with a slash."""
        if not self.package_href:
            if tag == "a" and not attrs[0][1].endswith("/"):
                self.package_href = attrs[0][1]
        else:
            super().handle_starttag(tag, attrs)


class PublishTestCase(unittest.TestCase):
    """Publish repositories with the rpm plugin."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)

        delete_orphans()

    def parse_date_from_string(self, s, parse_format="%Y-%m-%dT%H:%M:%S.%fZ"):
        """Parse string to datetime object.

        :param s: str like '2018-11-18T21:03:32.493697Z'
        :param parse_format: str defaults to %Y-%m-%dT%H:%M:%S.%fZ
        :return: datetime.datetime
        """
        return datetime.strptime(s, parse_format)

    def rpm_publish(self, url=RPM_KICKSTART_FIXTURE_URL, policy="on_demand"):
        """Publish repositories with the rpm plugin.

        This test targets the following issue:
        `Pulp #5630 <https://pulp.plan.io/issues/5630>`_.

        In order to sync a repository a remote has to be associated within
        this repository. When a repository is created this version field is set
        as None. After a sync the repository version is updated.

        Do the following:

        1. Create a repository and a remote.
        2. Assert that repository version is None.
        3. Sync the remote.
        4. Assert that repository version is not None.
        5. Assert that distribution_tree units were added and are present in the repo.
        6. Publish
        """
        delete_orphans()
        repo = self.client.post(RPM_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        # Create a remote with the standard test fixture url.
        body = gen_rpm_remote(url=url, policy=policy)
        remote = self.client.post(RPM_REMOTE_PATH, body)
        self.addCleanup(self.client.delete, remote["pulp_href"])

        # Sync the repository.
        self.assertEqual(repo["latest_version_href"], f"{repo['pulp_href']}versions/0/")
        data = {"remote": remote["pulp_href"]}
        response = self.client.using_handler(api.json_handler).post(
            urljoin(repo["pulp_href"], "sync/"), data
        )
        sync_task = self.client.get(response["task"])
        created_at = self.parse_date_from_string(sync_task["pulp_created"])
        started_at = self.parse_date_from_string(sync_task["started_at"])
        finished_at = self.parse_date_from_string(sync_task["finished_at"])
        task_duration = finished_at - started_at
        waiting_time = started_at - created_at
        print(
            "\n->     Sync => Waiting time (s): {wait} | Service time (s): {service}".format(
                wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
            )
        )

        repo = self.client.get(repo["pulp_href"])

        # Check that we have the correct content counts.
        self.assertIsNotNone(repo["latest_version_href"])

        self.assertIn(
            RPM_PACKAGE_CONTENT_NAME,
            get_content_summary(repo).keys(),
        )
        self.assertIn(
            RPM_PACKAGE_CONTENT_NAME,
            get_added_content_summary(repo).keys(),
        )

        repo = self.client.get(repo["pulp_href"])

        # Publishing
        body = {"repository": repo["pulp_href"]}
        response = self.client.using_handler(api.json_handler).post(RPM_PUBLICATION_PATH, body)
        publish_task = self.client.get(response["task"])
        created_at = self.parse_date_from_string(publish_task["pulp_created"])
        started_at = self.parse_date_from_string(publish_task["started_at"])
        finished_at = self.parse_date_from_string(publish_task["finished_at"])
        task_duration = finished_at - started_at
        waiting_time = started_at - created_at
        print(
            "\n->     Publish => Waiting time (s): {wait} | Service time (s): {service}".format(
                wait=waiting_time.total_seconds(), service=task_duration.total_seconds()
            )
        )
        return publish_task["created_resources"][0]

    def test_epel7(self):
        """Publish EPEL 7."""
        self.rpm_publish(url=EPEL7_URL)

    def test_centos7(self):
        """Publish CentOS 7."""
        self.rpm_publish(url=CENTOS7_URL)

    def test_centos_8stream_baseos(self):
        """Publish CentOS 8 BaseOS."""
        publication_href = self.rpm_publish(url=CENTOS8_STREAM_BASEOS_URL)
        # Test that the .treeinfo file is available and AppStream sub-repo is published correctly
        body = {"publication": publication_href, "name": "centos8", "base_path": "centos8"}
        response = self.client.using_handler(api.json_handler).post(RPM_DISTRIBUTION_PATH, body)
        distribution_task = self.client.get(response["task"])
        distribution_href = distribution_task["created_resources"][0]
        self.addCleanup(self.client.delete, distribution_href)
        distribution = self.client.get(distribution_href)
        treeinfo_file = utils.http_get(urljoin(distribution["base_url"], ".treeinfo"))
        treeinfo = TreeInfo()
        with NamedTemporaryFile("wb") as temp_file:
            temp_file.write(treeinfo_file)
            temp_file.flush()
            treeinfo.load(f=temp_file.name)
        for variant_name, variant in treeinfo.variants.variants.items():
            if variant_name == "BaseOS":
                self.assertEqual(variant.paths.repository, ".")
                self.assertEqual(variant.paths.packages, "Packages")
            elif variant_name == "AppStream":
                self.assertEqual(variant.paths.repository, "AppStream")
                self.assertEqual(variant.paths.packages, "AppStream/Packages")
                # Find the first package in the 'AppStream/Packages/a/' directory and download it
                parser = PackagesHtmlParser()
                a_packages_href = urljoin(
                    distribution["base_url"], "{}/a/".format(variant.paths.packages)
                )
                a_packages_listing = utils.http_get(a_packages_href)
                parser.feed(a_packages_listing.__str__())
                full_package_path = urljoin(a_packages_href, parser.package_href)
                utils.http_get(full_package_path)

    def test_centos8_appstream(self):
        """Publish CentOS 8 AppStream."""
        self.rpm_publish(url=CENTOS8_STREAM_APPSTREAM_URL)

    def test_centos8_baseos(self):
        """Publish CentOS 8 BaseOS."""
        self.rpm_publish(url=CENTOS8_STREAM_BASEOS_URL)



from functools import reduce
from logging import getLogger

from django.db import models
from django.db.models.signals import post_delete
from django.dispatch import receiver

from pulpcore.plugin.models import (
    Artifact,
    BaseModel,
    Content,
    ContentArtifact,
    Repository,
)

log = getLogger(__name__)


class DistributionTree(Content):
    """
    Model for an RPM distribution tree (also sometimes referenced as an installable tree).

    A distribution tree is described by a file in root of an RPM repository named either
    "treeinfo" or ".treeinfo". This INI file is used by system installers to boot from a URL.
    It describes the operating system or product contained in the distribution tree and
    where the bootable media is located for various platforms (where platform means
    'x86_64', 'xen', or similar).

    The description of the "treeinfo" format is included below, originally take from
    https://release-engineering.github.io/productmd/treeinfo-1.0.html

    Fields:
        header_version (Text):
            Metadata version
        release_name (Text):
            Release name
        release_short (Text):
            Release short name
        release_version (Text):
            Release version
        release_type (Text):
            Release type
        release_is_layered (Bool):
            Typically False for an operating system, True otherwise
        base_product_name (Text):
            Base product name
        base_product_short (Text):
            Base product short name
        base_product_version (Text):
            Base product *major* version
        base_product_type (Text):
            Base product release type
        arch (Text):
            Tree architecture
        build_timestamp (Float):
            Tree build time timestamp
        instimage (Text):
            Relative path to Anaconda instimage
        mainimage (Text):
            Relative path to Anaconda stage2 image
        discnum (Integer):
            Disc number
        totaldiscs (Integer):
            Number of discs in media set
        digest (Text):
            sha256 digest of the [.]treeinfo file

    """

    TYPE = "distribution_tree"

    header_version = models.TextField()

    release_name = models.TextField()
    release_short = models.TextField()
    release_version = models.TextField()
    release_is_layered = models.BooleanField(default=False)

    base_product_name = models.TextField(null=True)
    base_product_short = models.TextField(null=True)
    base_product_version = models.TextField(null=True)

    # tree
    arch = models.TextField()
    build_timestamp = models.FloatField()

    # stage2
    instimage = models.TextField(null=True)
    mainimage = models.TextField(null=True)

    # media
    discnum = models.IntegerField(null=True)
    totaldiscs = models.IntegerField(null=True)

    digest = models.TextField(null=False)

    def repositories(self):
        """
        Return the subrepos in this DistributionTree.

        Returns:
            django.db.models.QuerySet: Repos that are contained within this tree.

        """
        from pulp_rpm.app.models import RpmRepository

        repo_ids = list(self.addons.values_list("repository__pk", flat=True))
        repo_ids += list(self.variants.values_list("repository__pk", flat=True))
        return RpmRepository.objects.filter(pk__in=repo_ids)

    def content(self):
        """
        Return the subrepo content for this DistributionTree.

        Returns:
            django.db.models.QuerySet: Content that is contained within this tree.

        """
        if not self.repositories():
            return []

        version_content = [repo.latest_version().content for repo in self.repositories()]
        return reduce(lambda v1, v2: v1 | v2, version_content)

    def artifacts(self):
        """
        Return the subrepo artifacts for this DistributionTree.

        Returns:
            django.db.models.QuerySet: Content that is contained within this tree.

        """
        return Artifact.objects.filter(content__pk__in=self.content())

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = ("digest",)


class Checksum(BaseModel):
    """
    Distribution Tree Checksum.

    Checksums of selected files in a tree.

    Fields:
        path (Text):
            File path
        checksum (Text):
            Checksum value for the file

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree

    """

    path = models.TextField()
    checksum = models.TextField()
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name="checksums"
    )

    class Meta:
        unique_together = (
            "path",
            "checksum",
            "distribution_tree",
        )


class Image(BaseModel):
    """
    Distribution Tree Image.

    Images compatible with particular platform.

    Fields:
        name (Text):
            File name
        path (Text):
            File path
        platforms (Text):
            Compatible platforms

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree

    """

    name = models.TextField()
    path = models.TextField()
    platforms = models.TextField()
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name="images"
    )

    @property
    def artifact(self):
        """
        Returns artifact object.
        """
        content_artifact = ContentArtifact.objects.filter(
            content=self.distribution_tree,
            relative_path=self.path,
        ).first()

        artifact = content_artifact.artifact if content_artifact else None

        return artifact

    class Meta:
        unique_together = (
            "name",
            "path",
            "platforms",
            "distribution_tree",
        )


class Addon(BaseModel):
    """
    Distribution Tree Addon.

    Kickstart functionality expansion.

    Fields:
        addon_id (Text):
            Addon id
        uid (Text):
            Addon uid
        name (Text):
            Addon name
        type (Text):
            Addon type
        packages (Text):
            Relative path to directory with binary RPMs

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree
        repository (models.ForeignKey): The associated Repository

    """

    addon_id = models.TextField()
    uid = models.TextField()
    name = models.TextField()
    type = models.TextField()
    packages = models.TextField()
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name="addons"
    )
    repository = models.ForeignKey(Repository, on_delete=models.PROTECT, related_name="addons")

    class Meta:
        unique_together = (
            "addon_id",
            "uid",
            "name",
            "type",
            "packages",
            "distribution_tree",
        )


class Variant(BaseModel):
    """
    Distribution Tree Variant.

    Fields:
        variant_id (Text):
            Variant id
        uid (Text):
            Variant uid
        name (Text):
            Variant name
        type (Text):
            Variant type
        packages (Text):
            Relative path to directory with binary RPMs
        source_packages (Text):
            Relative path to directory with source RPMs
        source_repository (Text):
            Relative path to YUM repository with source RPMs
        debug_packages (Text):
            Relative path to directory with debug RPMs
        debug_repository (Text):
            Relative path to YUM repository with debug RPMs
        identity (Text):
            Relative path to a pem file that identifies a product

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree
        repository (models.ForeignKey): The associated Repository. If it's a main repo ('.'),
                                        it's not set and a reference to a repository a distribution
                                        tree belongs to is implied.

    """

    variant_id = models.TextField()
    uid = models.TextField()
    name = models.TextField()
    type = models.TextField()
    packages = models.TextField()
    source_packages = models.TextField(null=True)
    source_repository = models.TextField(null=True)
    debug_packages = models.TextField(null=True)
    debug_repository = models.TextField(null=True)
    identity = models.TextField(null=True)
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name="variants"
    )
    repository = models.ForeignKey(
        Repository, on_delete=models.PROTECT, related_name="variants", null=True
    )

    class Meta:
        unique_together = (
            "variant_id",
            "uid",
            "name",
            "type",
            "packages",
            "distribution_tree",
        )


@receiver(post_delete, sender=Addon)
@receiver(post_delete, sender=Variant)
def cleanup_subrepos(sender, instance, **kwargs):
    """
    Remove subrepos when a DistributionTree is being removed.

    """
    subrepo = None
    try:
        subrepo = instance.repository
    except Repository.DoesNotExist:
        pass
    if subrepo:
        Variant.objects.filter(repository=subrepo).delete()
        Addon.objects.filter(repository=subrepo).delete()
        subrepo.delete()



from gettext import gettext as _
import logging

from drf_spectacular.utils import extend_schema
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.serializers import ValidationError as DRFValidationError

from pulpcore.plugin.actions import ModifyRepositoryActionMixin
from pulpcore.plugin.models import RepositoryVersion
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.serializers import (
    AsyncOperationResponseSerializer,
)
from pulpcore.plugin.viewsets import (
    DistributionViewSet,
    NamedModelViewSet,
    OperationPostponedResponse,
    PublicationViewSet,
    RemoteViewSet,
    RepositoryVersionViewSet,
    RepositoryViewSet,
)

from pulp_rpm.app import tasks
from pulp_rpm.app.constants import SYNC_POLICIES
from pulp_rpm.app.models import (
    RpmDistribution,
    RpmPublication,
    RpmRemote,
    RpmRepository,
    UlnRemote,
)
from pulp_rpm.app.serializers import (
    CopySerializer,
    RpmDistributionSerializer,
    RpmPublicationSerializer,
    RpmRemoteSerializer,
    RpmRepositorySerializer,
    RpmRepositorySyncURLSerializer,
    UlnRemoteSerializer,
)


class RpmRepositoryViewSet(RepositoryViewSet, ModifyRepositoryActionMixin):
    """
    A ViewSet for RpmRepository.
    """

    endpoint_name = "rpm"
    queryset = RpmRepository.objects.exclude(user_hidden=True)
    serializer_class = RpmRepositorySerializer

    @extend_schema(
        description="Trigger an asynchronous task to sync RPM content.",
        summary="Sync from remote",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=RpmRepositorySyncURLSerializer)
    def sync(self, request, pk):
        """
        Dispatches a sync task.
        """
        repository = self.get_object()
        serializer = RpmRepositorySyncURLSerializer(
            data=request.data, context={"request": request, "repository_pk": pk}
        )
        serializer.is_valid(raise_exception=True)
        remote = serializer.validated_data.get("remote", repository.remote)
        mirror = serializer.validated_data.get("mirror")
        sync_policy = serializer.validated_data.get("sync_policy")
        skip_types = serializer.validated_data.get("skip_types")
        optimize = serializer.validated_data.get("optimize")

        if not sync_policy:
            sync_policy = SYNC_POLICIES.ADDITIVE if not mirror else SYNC_POLICIES.MIRROR_COMPLETE

        # validate some invariants that involve repository-wide settings.
        if sync_policy in (SYNC_POLICIES.MIRROR_COMPLETE, SYNC_POLICIES.MIRROR_CONTENT_ONLY):
            err_msg = (
                "Cannot use '{}' in combination with a 'mirror_complete' or "
                "'mirror_content_only' sync policy."
            )
            if repository.retain_package_versions > 0:
                raise DRFValidationError(err_msg.format("retain_package_versions"))

        if sync_policy == SYNC_POLICIES.MIRROR_COMPLETE:
            err_msg = "Cannot use '{}' in combination with a 'mirror_complete' sync policy."
            if repository.autopublish:
                raise DRFValidationError(err_msg.format("autopublish"))
            if skip_types:
                raise DRFValidationError(err_msg.format("skip_types"))

        result = dispatch(
            tasks.synchronize,
            shared_resources=[remote],
            exclusive_resources=[repository],
            kwargs={
                "sync_policy": sync_policy,
                "remote_pk": str(remote.pk),
                "repository_pk": str(repository.pk),
                "skip_types": skip_types,
                "optimize": optimize,
            },
        )
        return OperationPostponedResponse(result, request)


class RpmRepositoryVersionViewSet(RepositoryVersionViewSet):
    """
    RpmRepositoryVersion represents a single rpm repository version.
    """

    parent_viewset = RpmRepositoryViewSet


class RpmRemoteViewSet(RemoteViewSet):
    """
    A ViewSet for RpmRemote.
    """

    endpoint_name = "rpm"
    queryset = RpmRemote.objects.all()
    serializer_class = RpmRemoteSerializer


class UlnRemoteViewSet(RemoteViewSet):
    """
    A ViewSet for UlnRemote.
    """

    endpoint_name = "uln"
    queryset = UlnRemote.objects.all()
    serializer_class = UlnRemoteSerializer


class RpmPublicationViewSet(PublicationViewSet):
    """
    ViewSet for Rpm Publications.
    """

    endpoint_name = "rpm"
    queryset = RpmPublication.objects.exclude(complete=False)
    serializer_class = RpmPublicationSerializer

    @extend_schema(
        description="Trigger an asynchronous task to create a new RPM content publication.",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        """
        Dispatches a publish task.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        repository_version = serializer.validated_data.get("repository_version")
        repository = RpmRepository.objects.get(pk=repository_version.repository.pk)

        metadata_checksum_type = serializer.validated_data.get(
            "metadata_checksum_type", repository.metadata_checksum_type
        )
        package_checksum_type = serializer.validated_data.get(
            "package_checksum_type", repository.package_checksum_type
        )
        checksum_types = dict(
            metadata=metadata_checksum_type,
            package=package_checksum_type,
        )
        gpgcheck_options = dict(
            gpgcheck=serializer.validated_data.get("gpgcheck", repository.gpgcheck),
            repo_gpgcheck=serializer.validated_data.get("repo_gpgcheck", repository.repo_gpgcheck),
        )
        sqlite_metadata = serializer.validated_data.get(
            "sqlite_metadata", repository.sqlite_metadata
        )
        if sqlite_metadata:
            logging.getLogger("pulp_rpm.deprecation").info(
                "Support for sqlite metadata generation will be removed from a future release "
                "of pulp_rpm. See https://tinyurl.com/sqlite-removal for more details"
            )

        if repository.metadata_signing_service:
            signing_service_pk = repository.metadata_signing_service.pk
        else:
            signing_service_pk = None

        result = dispatch(
            tasks.publish,
            shared_resources=[repository_version.repository],
            kwargs={
                "repository_version_pk": repository_version.pk,
                "metadata_signing_service": signing_service_pk,
                "checksum_types": checksum_types,
                "gpgcheck_options": gpgcheck_options,
                "sqlite_metadata": sqlite_metadata,
            },
        )
        return OperationPostponedResponse(result, request)


class RpmDistributionViewSet(DistributionViewSet):
    """
    ViewSet for RPM Distributions.
    """

    endpoint_name = "rpm"
    queryset = RpmDistribution.objects.all()
    serializer_class = RpmDistributionSerializer


class CopyViewSet(viewsets.ViewSet):
    """
    ViewSet for Content Copy.
    """

    serializer_class = CopySerializer

    @extend_schema(
        description="Trigger an asynchronous task to copy RPM content"
        "from one repository into another, creating a new"
        "repository version.",
        summary="Copy content",
        operation_id="copy_content",
        request=CopySerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        """Copy content."""
        serializer = CopySerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)

        dependency_solving = serializer.validated_data["dependency_solving"]
        config = serializer.validated_data["config"]

        config, shared_repos, exclusive_repos = self._process_config(config)

        async_result = dispatch(
            tasks.copy_content,
            shared_resources=shared_repos,
            exclusive_resources=exclusive_repos,
            args=[config, dependency_solving],
            kwargs={},
        )
        return OperationPostponedResponse(async_result, request)

    def _process_config(self, config):
        """
        Change the hrefs into pks within config.

        This method also implicitly validates that the hrefs map to objects and it returns a list of
        repos so that the task can lock on them.
        """
        result = []
        # exclusive use of the destination repos is needed since new repository versions are being
        # created, but source repos can be accessed in a read-only fashion in parallel, so long
        # as there are no simultaneous modifications.
        shared_repos = []
        exclusive_repos = []

        for entry in config:
            r = dict()
            source_version = NamedModelViewSet().get_resource(
                entry["source_repo_version"], RepositoryVersion
            )
            dest_repo = NamedModelViewSet().get_resource(entry["dest_repo"], RpmRepository)
            r["source_repo_version"] = source_version.pk
            r["dest_repo"] = dest_repo.pk
            shared_repos.append(source_version.repository)
            exclusive_repos.append(dest_repo)

            if "dest_base_version" in entry:
                try:
                    r["dest_base_version"] = dest_repo.versions.get(
                        number=entry["dest_base_version"]
                    ).pk
                except RepositoryVersion.DoesNotExist:
                    message = _(
                        "Version {version} does not exist for repository " "'{repo}'."
                    ).format(version=entry["dest_base_version"], repo=dest_repo.name)
                    raise DRFValidationError(detail=message)

            if entry.get("content") is not None:
                r["content"] = []
                for c in entry["content"]:
                    r["content"].append(NamedModelViewSet().extract_pk(c))
            result.append(r)

        return result, shared_repos, exclusive_repos



import requests


from pulp_smash.pulp3.bindings import (
    delete_orphans,
    monitor_task,
    PulpTestCase,
)
from pulp_smash.pulp3.utils import (
    gen_distribution,
    gen_repo,
)

from pulp_rpm.tests.functional.utils import gen_rpm_client

from pulpcore.client.pulp_rpm import (
    DistributionsRpmApi,
    RepositoriesRpmApi,
    PublicationsRpmApi,
    RpmRpmPublication,
)


class ContentHandlerTests(PulpTestCase):
    """Whether the RpmDistribution.content_handler* methods work."""

    @classmethod
    def tearDownClass(cls) -> None:
        """Clean up after testing."""
        delete_orphans()

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the class."""
        cls.client = gen_rpm_client()
        cls.repo_api = RepositoriesRpmApi(cls.client)
        cls.publications_api = PublicationsRpmApi(cls.client)
        cls.distributions_api = DistributionsRpmApi(cls.client)
        delete_orphans()

    def setUp(self):
        """Setup method."""
        self.repo = self.repo_api.create(gen_repo())

        publish_data = RpmRpmPublication(repository=self.repo.pulp_href)
        publish_response = self.publications_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        self.publication = self.publications_api.read(created_resources[0])

        dist_data = gen_distribution(publication=self.publication.pulp_href)
        dist_response = self.distributions_api.create(dist_data)
        created_resources = monitor_task(dist_response.task).created_resources
        self.dist = self.distributions_api.read(created_resources[0])

    def tearDown(self):
        """Tearddown method."""
        if self.publication:
            self.publications_api.delete(self.publication.pulp_href)
        self.repo_api.delete(self.repo.pulp_href)
        self.distributions_api.delete(self.dist.pulp_href)

    def test_config_repo_in_listing_unsigned(self):
        """Whether the served resources are in the directory listing."""
        resp = requests.get(self.dist.base_url)

        self.assertEqual(resp.status_code, 200)
        self.assertIn(b"config.repo", resp.content)
        self.assertNotIn(b"repomd.xml.key", resp.content)

    def test_config_repo_unsigned(self):
        """Whether config.repo can be downloaded and has the right content."""
        self.config_repo_check()

    def test_config_repo_auto_distribute(self):
        """Whether config.repo is properly served using auto-distribute."""
        publication_href = self.dist.publication
        body = {"repository": self.repo.pulp_href, "publication": None}
        monitor_task(self.distributions_api.partial_update(self.dist.pulp_href, body).task)
        # Check that distribution is now using repository to auto-distribute
        self.dist = self.distributions_api.read(self.dist.pulp_href)
        self.assertEqual(self.repo.pulp_href, self.dist.repository)
        self.assertIsNone(self.dist.publication)
        self.config_repo_check()
        # Delete publication and check that 404 is now returned
        self.publications_api.delete(publication_href)
        self.publication = None
        resp = requests.get(f"{self.dist.base_url}config.repo")
        self.assertEqual(resp.status_code, 404)

    def config_repo_check(self):
        """Helper to do the tests on config.repo."""
        resp = requests.get(f"{self.dist.base_url}config.repo")

        self.assertEqual(resp.status_code, 200)
        self.assertIn(bytes(f"[{self.dist.name}]\n", "utf-8"), resp.content)
        self.assertIn(bytes(f"baseurl={self.dist.base_url}\n", "utf-8"), resp.content)
        self.assertIn(bytes("gpgcheck=0\n", "utf-8"), resp.content)
        self.assertIn(bytes("repo_gpgcheck=0", "utf-8"), resp.content)
